;; Generated package description from denote-search.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "denote-search" "1.0.2" "Search the contents of your notes" '((emacs "29.1") (denote "3.0")) :commit "02d2d6d361d5e86376b2a1b102515cc899efabd2" :authors '(("Lucas Quintana" . "lmq10@protonmail.com")) :maintainer '("Lucas Quintana" . "lmq10@protonmail.com") :keywords '("matching") :url "https://github.com/lmq-10/denote-search")
